"""
Models for querying Rune users.

"""
from .common import SetMemberBase



class Org(SetMemberBase):
    """
    Representational state for a Rune organization.

    """


class Membership(SetMemberBase):
    """
    Representational state for a user's membership to an organization.

    """

    _relations = {
        'org': Org
    }


class User(SetMemberBase):
    """
    Representational state for a Rune platform user.

    """

    _relations = {
        'defaultMembership': Membership
    }



"""
Common utilities and base classes.

"""
from collections import OrderedDict
import re
from typing import Any, Dict, Iterable, Iterator, Optional

from .graph.model import GraphID


INDENTATION = "    "
"""
Indentation string used in formatting.

"""

_camel_to_snake = re.compile(r'(?<!^)(?=[A-Z])')
"""
Regular expression pattern for coercing CamelCase to snake_case.

"""


class SetMemberBase:
    """
    Base class for graph models representing a single member of a set.

    """

    _id: Optional[GraphID] = None
    """
    The Graph item ID.

    """

    _relations: Dict = {}
    """
    Mapping of fields to model classes.

    """


    def __eq__(self, right):
        """
        Compare two graph items for equality by ID.

        """
        if isinstance(right, SetMemberBase):
            return self._id == right._id
        else:
            return self._id == right


    def __getattr__(self, attr: str) -> Any:
        """
        Facilitate access to fields in both canonical snake case and their
        original GraphQL case (lowercase camel).

        """
        if '_' in attr:
            field_name = attr[0] + attr.title()[1:].replace('_', '')
        else:
            field_name = attr

        try:
            return self._res[field_name]
        except KeyError:
            raise AttributeError(field_name)


    def __getitem__(self, field_name: str) -> Any:
        """
        Access the raw representational state directly.

        """
        return self._res[field_name]


    def __init__(self, res: Dict[str, Any], session=None):
        """
        Initialize representational state.

        res
            The raw representational state dictionary, as returned by the
            Graph API. Note that this object takes ownership, and may modify
            it in-place. The caller should not continue to use the state.

        session
            Shared reference to the client session.

        """
        try:
            self._id = GraphID(res['id'], self.resource)
        except KeyError:
            pass
        else:
            res['id'] = self._id.unqualified

        for field_name, model in self._relations.items():
            try:
                field_res = res[field_name]
            except KeyError:
                continue

            if isinstance(field_res, list):
                res[field_name] = [model(state) for state in field_res]
            else:
                res[field_name] = model(field_res)

        self._res = res
        self._session = session


    def __iter__(self):
        """
        Iterator over the representational state dictionary.

        """
        return iter(self._res)


    def __len__(self):
        """
        Number of fields in state.

        """
        return len(self._res)


    def __repr__(self):
        """
        Format the diagnostic representaion.

        """
        return repr(self._res)


    def __str__(self):
        """
        Format the human-readable representation.

        """
        return self._format(0)


    def _format(self, indent: int) -> str:
        """
        Format human-readable string representation with indentation.

        """
        fmt = f"{self.resource} {{\n"

        for field, value in self._res.items():
            fmt += INDENTATION * (indent + 1)
            attr = _camel_to_snake.sub('_', field).lower()

            if isinstance(value, SetMemberBase):
                # Recursively format child
                fmt += f"{attr}: {value._format(indent + 1)}\n"

            elif (
                isinstance(value, list)
                and
                len(value)
                and isinstance(value[0], SetMemberBase)
            ):
                # Recursively format many children
                fmt += f"{attr}: [\n"

                for child in value:
                    fmt += INDENTATION * (indent + 2)
                    fmt += f"{child._format(indent + 2)}\n"

                fmt += INDENTATION * (indent + 1)
                fmt += "]\n"

            else:
                # Scalar
                fmt += f"{attr}: {value}\n"

        fmt += INDENTATION * indent
        fmt += '}'

        return fmt


    @property
    def resource(self) -> str:
        """
        The resource name.

        """
        return self.__class__.__name__.lower()


class SetBase:
    """
    Base class for collections of items.

    """

    _all: bool
    """
    Metadata about whether this contains all the available members of the
    set.

    """

    _member: SetMemberBase
    """
    The collection member class.

    """

    _members: OrderedDict
    """
    The members of this set.

    """


    def __getitem__(self, key: str):
        """
        Get a member by ID.

        """
        if isinstance(key, str):
            key = GraphID(key)

        return self._members[key.as_tuple()]


    def __init__(self, all_=False):
        """
        Initialize the base

        """
        self._all = all_
        self._members = OrderedDict()


    def __iter__(self) -> Iterator:
        """
        Iterate over the members of the set.

        """
        return iter(self._members.values())


    def __str__(self):
        """
        Format the list of members for human readability.

        """
        return "\n".join(str(member) for member in self._members.values())


    def add(self, member):
        """
        Add a member to this set.

        """
        self._members[member._id.as_tuple()] = member


    def created_before(self, t: float) -> Iterator:
        """
        Iterator for members created before a given time.

        """
        for member in self._members.values():
            t_created = member.created_at

            if t_created is not None and member.created_at >= t:
                continue

            yield member


    def filtered_by(self, **conditions) -> Iterator:
        """
        Filter members with specific attributes equal to a value.

        Return an iterator over only items that match ALL of the attribute
        constraints. This is identical to `find_all_by()` on a query, but for
        cached sets that are already loaded.

        ```
        # Get all apple watches from a cached device set
        for device in device_cache.filtered_by(device_type='apple watch'):
            do_something(device)
        ```

        This method is intended for internal SDK use only.

        """
        if not conditions:
            raise TypeError(
                "at least one attribute filter condition expected")

        for member in self._members.values():

            for attr, value in conditions.items():
                if getattr(member, attr) != value:
                    break
            else:
                yield member


    def ids(self) -> Iterator[str]:
        """
        Iterator over the member IDs.

        """
        return (member.id for member in self._members.values())


    def update(self, other: Iterable):
        """
        Add members to the set from an iterable

        """
        for member in other:
            self.add(member)


class QueryBase:
    """
    Base class for query classes.

    """

    _set: SetBase
    """
    The set class.

    """


    def __init__(self, session):
        """
        Initialize the base

        """
        self._session = session


    def __getitem__(self, member_id: str):
        """
        Retrieve a specific member of the collection by its ID.

        """
        raise NotImplementedError()


    def __iter__(self):
        """
        Execute a streaming query and return an iterator over the results.

        Uses the local cache when enabled.

        """
        raise NotImplementedError()


    def __str__(self):
        """
        Execute the query.

        This magic behavior is implemented to support, e.g.
        `print(rune.patients)`.

        """
        result = self._set(all_=True)

        for member in self:
            result.add(member)

        return str(result)


    def find_all_by(self, **conditions) -> Iterator:
        """
        Find subset with specific attributes.

        """
        if not conditions:
            raise TypeError(
                "at least one attribute filter condition expected")

        for member in self:

            for attr, value in conditions.items():
                if getattr(member, attr) != value:
                    break
            else:
                yield member


    def query(self):
        """
        Execute query and return results.

        This method always executes an API query. Use `iter(query)` for
        general use instead.

        """
        sess = self._session
        result = self._set(all_=True)

        for member_state in self.raw_query():
            result.add(self._set._member(member_state, session=sess))

        return result


    def raw_query(self) -> Iterable:
        """
        Execute a GraphQL query to the API and return the raw state result.

        """
        raise NotImplementedError()




"""
Tests for the v2 client.

"""
from unittest import TestCase
from unittest.mock import patch

import runeq
from runeq.config import AUTH_METHOD_ACCESS_TOKEN, AUTH_METHOD_CLIENT_KEYS
from runeq.v2sdk.patients import Patient
from runeq.v2sdk.users import User


class TestClientSession(TestCase):
    """
    Unit tests for the client session.

    """


    @patch('runeq.v2sdk.client.GraphAPIClient')
    def test_i_patient(self, graphclient):
        """
        Test the "who am I?" functionality as a patient.

        """
        config = runeq.Config(
            auth_method=AUTH_METHOD_CLIENT_KEYS,
            graph_url='http://graph.example.io',
            client_key_id='llama',
            client_access_key='llama',
        )

        graphclient().whoami.return_value = {
            'patient': {
                'id': 'patient-llama,patient',
                'codeName': 'llama code name',
                'createdAt': 1234567890.0,
            }
        }

        rune = runeq.session(config)
        me = rune.me

        self.assertIsInstance(me, Patient)
        self.assertEqual('llama', me.id)
        self.assertEqual('llama code name', me.code_name)
        graphclient().whoami.assert_called_once()

        # Assert caching works
        me = rune.me
        graphclient().whoami.assert_called_once()


    @patch('runeq.v2sdk.client.GraphAPIClient')
    def test_i_user(self, graphclient):
        """
        Test the "who am I" functionality as a user.

        """
        config = runeq.Config(
            auth_method=AUTH_METHOD_ACCESS_TOKEN,
            graph_url='http://graph.example.io',
            access_token_id='llama',
            access_token_secret='llama'
        )

        graphclient().whoami.return_value = {
            'user': {
                'id': 'llama',
                'created': 1234567890.0,
                'defaultMembership': {
                    'id': 'rune$llama',
                    'created': 1234567890.0,
                    'org': {
                        'id': 'llama',
                        'created': 1234567890.0,
                        'displayName': 'Llama Org'
                    }
                },
                'displayName': 'Llama User',
                'email': 'llama@runelabs.io',
                'username': 'llamauser',
            }
        }

        rune = runeq.session(config)
        me = rune.me

        self.assertIsInstance(me, User)
        self.assertEqual('llama', me.id)
        self.assertEqual('llama@runelabs.io', me.email)
        graphclient().whoami.assert_called_once()

        # Assert caching works
        me = rune.me
        graphclient().whoami.assert_called_once()



"""
Client for the Rune Labs graph API.

This package is not intended to be used directly, and implementation is
subject to change. Use the functionality exposed by the `runeq.session()`
client session instead.

"""



"""
The Rune Labs Standard Development Kit (SDK) for querying data from Rune
APIs.

"""
from . import stream
from .config import Config
from .v2sdk import session


__all__ = [
    'Config',
    'session',
    'stream',
]



"""
Models for querying device metadata.

"""
from typing import Iterable, Iterator, Optional, Union

from .common import SetMemberBase, QueryBase, SetBase
from .graph.model import GraphID


class DeviceType(SetMemberBase):
    """
    Representational state for a device type.

    """


    def __eq__(self, right):
        """
        Case-insensitive device type name comparison.

        """
        if isinstance(right, str):
            right = right.lower()
            return (
                self.id.lower() == right
                or
                self.display_name.lower() == right
            )
        else:
            return super().__eq__(right)


class Device(SetMemberBase):
    """
    Representational state for a patient device.

    """

    _relations = {
        'deviceType': DeviceType
    }


class DeviceSet(SetBase):
    """
    Representational state for a collection of devices.

    """

    _member = Device
    """
    Type for a member of a device set.

    """

    _patient_id: Optional[GraphID] = None
    """
    The patient, if any, that all devices in this set belong to.

    """


    def __getitem__(self, device_id: str) -> Device:
        """
        Get a device by ID.

        """
        if ',' in device_id:
            # Absolute graph item ID
            key = GraphID(device_id).as_tuple()

        elif self._patient_id:
            # Relative device ID
            if '-' not in device_id:
                device_id = f"device-{device_id}"

            key = (self._patient_id.principal, device_id)

        else:
            raise KeyError(device_id)

        return super().__getitem__(key)


    def ids(self) -> Iterator[Union[str, GraphID]]:
        """
        Iterator over device IDs.

        """
        if self._patient_id:
            # Return relative IDs
            return (
                device.id for device in self._members.values()
            )
        else:
            # Return absolute IDs
            return (
                device._id for device in self._members.values()
            )


class DeviceQuery(QueryBase):
    """
    A query over devices for one or all patients.

    """

    _set = DeviceSet
    """
    Type for a set of devices.

    """


    def __getitem__(self, device_id: str) -> Device:
        """
        Fetch a device by ID.

        Only supported with a patient in context.

        """
        if self._patient:

            if ',' in device_id:
                # Qualified ID: parse and verify
                device_id = GraphID(device_id)

                if device_id.principal != self._patient._id.principal:
                    raise ValueError(
                        "device does not belong to the specified patient")

                device_id = device_id.unqualified

            elif '-' in device_id:
                # Strip resource prefix
                device_id = device_id.split('-', 1)[-1]

            # Load all patient devices and scan for it
            for device in self:
                if device.id == device_id:
                    return device

            raise KeyError(device_id)

        else:
            raise TypeError("no patient specified for device query")


    def __init__(self, session, patient=None):
        """
        Initialize a new query.

        """
        super().__init__(session=session)
        self._patient = patient


    def __iter__(self):
        """
        Iterate over the results.

        The patient and device cache will be used, if caching is enabled.

        """
        sess = self._session
        patient = self._patient

        if patient:
            #
            # Devices of one patient
            #
            if sess._caching:
                # Work from the cache
                if not patient._device_cache._all:
                    patient._device_cache = self.query()

                if sess._now is not None:
                    yield from patient._device_cache.created_before(sess._now)
                else:
                    yield from iter(patient._device_cache)

            else:
                # Stream
                for device_state in self.raw_query():
                    device = Device(device_state)

                    if (sess._now is not None
                            and device.created_at >= sess._now):
                        continue

                    yield device

        else:
            #
            # Iterate devices across all patients
            #
            for patient in sess.patients:
                yield from iter(patient.devices)


    def query(self):
        """
        Execute patient query and return results.

        This method always executes an API query. Use `iter(query)` for
        general use instead.

        """
        result = super().query()

        if self._patient:
            result._patient_id = self._patient._id

        return result


    def raw_query(self) -> Iterable:
        """
        Execute patient devices query and return results.

        This method always executes an API query. Use `iter(devices)` if
        you want to take advantage of caching and `session.freeze_time()`.

        """
        sess = self._session
        patient = self._patient

        if patient:
            #
            # Devices of one patient
            #
            yield from sess.graph.list_patient_devices(patient_id=patient._id)

        else:
            #
            # All devices across all patients
            #
            for patient in sess.patients:
                yield from patient.devices.raw_query()



"""
Base for Graph API representational state models.

"""
from typing import Optional, Tuple


class GraphID:
    """
    A fully qualified Graph item key.

    """

    _key: Tuple[str]
    """
    The key components.

    """


    def __contains__(self, sub):
        """
        Proxy substring queries.

        """
        if sub == ',':
            return len(self._key) > 1
        elif sub == '-':
            return True
        elif isinstance(sub, str):
            return sub in self._key[0] or sub in self._key[-1]
        else:
            return TypeError(f"in operator not supported for {type(sub)}")


    def __eq__(self, right):
        """
        Compare two graph IDs for equality.

        """
        if isinstance(right, GraphID):
            return self._key == right._key
        elif isinstance(right, str):
            return ','.join(self._key) == right
        else:
            return False


    def __init__(self, key, resource: Optional[str] = None):
        """
        Initialize the graph ID.

        """
        if isinstance(key, self.__class__):
            # Copy graph ID
            self._key = key._key
            return

        if ',' in key:
            self._key = tuple(key.split(',', 1))

        elif resource:

            if '-' not in key:
                key = f"{resource}-{key}"

            self._key = (key, resource)

        else:
            raise ValueError(f"ambiguous graph key: {key!r}")


    def __str__(self):
        """
        Format back to string Graph identifier.

        """
        return ",".join(self._key)


    def as_tuple(self):
        """
        Return the single or double containing the Graph item primary key.

        """
        return self._key


    @property
    def principal(self):
        """
        Return the principal component of the key.

        """
        return self._key[0]


    @property
    def relative(self):
        """
        Return the relative component of the key.

        """
        if len(self._key) > 1:
            return self._key[1]


    @property
    def unqualified(self):
        """
        The relative, unqualified identifier.

        """
        if len(self._key) > 1 and '-' in self._key[1]:
            return self._key[1].split('-', 1)[-1]
        else:
            return self._key[0].split('-', 1)[-1]



"""
Tests for patient queries and record interaction.

"""
from unittest import TestCase
from unittest.mock import MagicMock, patch

import runeq


class TestPatients(TestCase):
    """
    Tests for interacting with the patient graph resource.

    """


    @patch('runeq.v2sdk.graph.client.GQLClient')
    def test_fetch_patient_by_id(self, gqlclient):
        """
        Test listing all patients with a frozen time session.

        """
        session = runeq.session(MagicMock())
        gqlclient().execute.return_value = {
            'patient': {
                'id': 'patient-llama,patient',
                'codeName': 'Llama Patient',
                'createdAt': 1234567890.0
            }
        }

        patient = session.patients['llama']

        self.assertEqual('llama', patient.id)
        self.assertEqual('Llama Patient', patient.code_name)
        self.assertEqual(1234567890.0, patient.created_at)
        gqlclient().execute.assert_called_once()

        # Test patient caching
        patient = session.patients['llama']
        gqlclient().execute.assert_called_once()


    @patch('runeq.v2sdk.client.GraphAPIClient')
    def test_find_by_codename(self, graphclient):
        """
        Test searching for a patient by code name.

        """
        session = runeq.session(MagicMock())
        graphclient().list_patients.side_effect = lambda: iter(
            [
                {
                    'id': 'patient-llama0,patient',
                    'codeName': 'Llama Zero',
                    'createdAt': 1234567890.0
                },
                {
                    'id': 'patient-llama1,patient',
                    'codeName': 'Llama One',
                    'createdAt': 1234567890.1
                },
                {
                    'id': 'patient-llama2,patient',
                    'codeName': 'Llama Two',
                    'createdAt': 1234567890.2
                }
            ]
        )

        patients = list(session.patients.find_all_by(code_name='Llama One'))

        self.assertEqual(1, len(patients))
        self.assertEqual('llama1', patients[0].id)


    @patch('runeq.v2sdk.graph.client.GQLClient')
    def test_frozen_patients(self, gqlclient):
        """
        Test listing all patients with a frozen time session.

        """
        session = runeq.session(MagicMock())
        gqlclient().execute.return_value = {
            'org': {
                'patientAccessList': {
                    'pageInfo': {'endCursor': None},
                    'patientAccess': [
                        {
                            'patient': {
                                'id': 'patient-llama0,patient',
                                'codeName': 'Patient Llama 0',
                                'createdAt': 1234567890.0,
                            }
                        },
                        {
                            'patient': {
                                'id': 'patient-llama1,patient',
                                'codeName': 'Patient Llama 1',
                                'createdAt': 1234567890.1,
                            }
                        }
                    ]
                }
            }
        }

        session.freeze_time(1234567890.05)
        patients = list(session.patients)

        self.assertEqual(1, len(patients))
        self.assertEqual('llama0', patients[0].id)
        gqlclient().execute.assert_called_once()


    @patch('runeq.v2sdk.graph.client.GQLClient')
    def test_list_all_patients(self, gqlclient):
        """
        Test listing all patients.

        """
        session = runeq.session(MagicMock())
        gqlclient().execute.return_value = {
            'org': {
                'patientAccessList': {
                    'pageInfo': {'endCursor': None},
                    'patientAccess': [
                        {
                            'patient': {
                                'id': 'patient-llama0,patient',
                                'codeName': 'Patient Llama 0',
                                'createdAt': 1234567890.0,
                            }
                        },
                        {
                            'patient': {
                                'id': 'patient-llama1,patient',
                                'codeName': 'Patient Llama 1',
                                'createdAt': 1234567890.1,
                            }
                        }
                    ]
                }
            }
        }

        patients = list(session.patients)

        self.assertEqual(2, len(patients))
        self.assertEqual('llama0', patients[0].id)
        self.assertEqual('llama1', patients[1].id)
        gqlclient().execute.assert_called_once()

        # Test that caching is working
        patients = list(session.patients)
        self.assertEqual(2, len(patients))
        self.assertEqual('llama0', patients[0].id)
        self.assertEqual('llama1', patients[1].id)
        gqlclient().execute.assert_called_once()

        # Test that retrieve uses the cache
        patient = session.patients['llama1']
        self.assertEqual('llama1', patient.id)
        gqlclient().execute.assert_called_once()



"""
Rune graph API client.

"""
import ast
from typing import Dict, Iterator, Optional

from gql import Client as GQLClient, gql
from gql.transport.requests import RequestsHTTPTransport

from runeq.config import AUTH_METHOD_CLIENT_KEYS, Config

from .model import GraphID


class RuneGraphAPIError(Exception):
    """
    Error from the Graph API.

    """


class RuneGraphNotFoundError(RuneGraphAPIError):
    """
    Graph API error indicating requested resource was not found.

    """


class GraphAPIClient:
    """
    Version 2 client for the Rune GraphQL API.

    """

    _config: Config
    """
    Reference to the SDK client configuration.

    """

    _gqlclient: GQLClient
    """
    The GraphQL client.

    """


    def __init__(self, config: Config):
        """
        Initialize the Graph API client.

        """
        self._config = config
        self._gqlclient = GQLClient(
            transport=RequestsHTTPTransport(
                retries=10,
                url=f"{config.graph_url}/graphql",
                use_json=True,
                headers={
                    'Content-Type': "application/json",
                    **config.auth_headers
                }
            ),
            fetch_schema_from_transport=True,
        )


    def fetch_patient(self, patient_id: GraphID) -> Optional[Dict]:
        """
        Attempt to retrieve a single patient by ID.

        """
        response = self.query(
            '''
            query($patientId: ID!) {
                patient(id: $patientId) {
                    id
                    codeName
                    createdAt
                }
            }
            ''',
            patientId=str(patient_id)
        )

        return response['patient']


    def list_patient_devices(self, patient_id: GraphID) -> Iterator[Dict]:
        """
        List all devices for a patient.

        """
        cursor = None

        while True:
            response = self.query(
                '''
                query(
                    $patientId: ID!,
                    $withDisabled: Boolean!,
                    $cursor: Cursor
                ) {
                    patient(id: $patientId) {
                        deviceList(
                            withDisabled: $withDisabled,
                            cursor: $cursor
                        ) {
                            devices {
                                id
                                alias
                                createdAt
                                deviceType {
                                    id
                                    displayName
                                }
                            }
                            pageInfo {
                                endCursor
                            }
                        }
                    }
                }
                ''',
                cursor=cursor,
                patientId=str(patient_id),
                withDisabled=False
            )
            result = response['patient']['deviceList']

            for device in result['devices']:
                device['patientId'] = patient_id.unqualified
                yield device

            cursor = result['pageInfo'].get('endCursor')

            if not cursor:
                break


    def list_patients(self) -> Iterator[Dict]:
        """
        List all patients this user has access to.

        """
        cursor = None

        while True:
            response = self.query(
                '''
                query($cursor: Cursor) {
                    org {
                        patientAccessList(cursor: $cursor) {
                            pageInfo {
                                endCursor
                            }
                            patientAccess {
                                patient {
                                    id
                                    codeName
                                    createdAt
                                }
                            }
                        }
                    }
                }
                ''',
                cursor=cursor
            )
            result = response['org']['patientAccessList']

            for patient_access in result['patientAccess']:
                yield patient_access['patient']

            cursor = result['pageInfo'].get('endCursor')

            if not cursor:
                break


    def query(self, statement: str, **variables) -> Dict:
        """
        Execute a GraphQL query against the API.

        """
        try:
            return self._gqlclient.execute(
                gql(statement),
                variable_values=variables
            )
        except Exception as error:
            #
            # Hack around GQL 2.0's lousy error reporting. Looks like there
            # may be a better client in the works.
            #
            try:
                detail = ast.literal_eval(error.args[0])
            except ValueError:
                raise RuneGraphAPIError(*error.args) from None

            if detail['extensions']['code'] == 'NotFoundError':
                raise RuneGraphNotFoundError() from None
            else:
                raise RuneGraphAPIError(detail['message']) from None


    def whoami(self) -> Dict:
        """
        Query the identity of either the patient or user whose credentials are
        configured.

        """
        if self._config.auth_method == AUTH_METHOD_CLIENT_KEYS:
            return self.query(
                '''
                query {
                    patient {
                        id
                        codeName
                        createdAt
                    }
                }
                '''
            )
        else:
            return self.query(
                '''
                query {
                    user {
                        id
                        created
                        defaultMembership {
                            id
                            created
                            org {
                                id
                                created
                                displayName
                            }
                        }
                        displayName
                        email
                        username
                    }
                }
                '''
            )



"""
Version 2 client for Rune's query APIs.

"""
from time import time
from typing import Optional, Union

from runeq import Config
from .devices import DeviceQuery
from .graph.client import GraphAPIClient
from .patients import PatientsQuery, Patient, PatientSet
from .users import User


class ClientSession:
    """
    Version 2 client for interacting with data on the Rune platform.

    """

    _caching: bool
    """
    Whether caching is enabled for this session.

    """

    _config: Config
    """
    The SDK configuration.

    """

    _graph: GraphAPIClient
    """
    Graph API client.

    """

    _me: Optional[Union[User, Patient]] = None
    """
    Cached caller identity.

    """

    _now: Optional[float] = None
    """
    Timestamp to use to limit query results instead of actual current time.

    """

    _patient_cache: PatientSet
    """
    Cached patient set.

    """


    def __init__(self, config: Config, caching=True):
        """
        Initialize the client.

        """
        self._caching = caching
        self._config = config
        self._graph = GraphAPIClient(config=config)
        self._patient_cache = PatientSet()


    @property
    def devices(self):
        """
        Begin a new query over devices across all patients.

        """
        return DeviceQuery(session=self)


    def freeze_time(self, at: Optional[float] = None):
        """
        Limit all queries to data available at the time of this call, and not
        after. This acts like an implicit end time for all your queries.

        Use this at the beginning of a multi-step analysis to prevent the
        possibility of new incoming data becoming available between subsequent
        queries. This can also increase performance in certain cases.

        For example, if you freeze at time T, you will consistently get the
        list of the same patients throughout your notebook, no matter how long
        it takes to run all the cells, and even if a new patient is registered
        at time T+1.

        at
            UNIX timestamp of time to limit results to. Default is now.

        """
        self._now = time() if at is None else at


    @property
    def graph(self) -> GraphAPIClient:
        """
        Reference to the graph client.

        """
        return self._graph


    @property
    def me(self) -> Union[User, Patient]:
        """
        Return information about the user identity whose credentials are
        configured.

        Information is loaded directly from the Rune API, and thus makes this
        a good utility for verifying your local configuration and API
        credentials.

        """
        if self._me is None:
            ident = self.graph.whoami()

            if 'patient' in ident:
                self._me = Patient(ident['patient'], session=self)
            else:
                self._me = User(ident['user'], session=self)

        return self._me


    @property
    def patients(self):
        """
        Begin a new query over patients accessible to this client.

        """
        return PatientsQuery(session=self)


    def unfreeze_time(self):
        """
        Remove the implicit time bound on results.

        """
        self._patient_cache = PatientSet()
        self._now = None




"""
Version 2 client for the Rune analysis API.

"""
from sys import stderr
from typing import Optional, Union

from runeq import Config
from .client import ClientSession


_beta_warned = False
"""
Whether the warning about beta usage has been printed.

"""


def beta_warning():
    """
    Issue a warning about using SDK code in beta testing.

    """
    global _beta_warned

    if not _beta_warned:
        print(
            "WARNING: You are accessing a part of the Rune SDK that is "
            "currently in Beta. Please be aware that Beta functionality is "
            "subject to change in future releases, and may break your code "
            "upon upgrade.",
            file=stderr,
        )
        _beta_warned = True


def session(
    config: Optional[Union[Config, str]] = None,
    caching=True,
    **kwarg,
):
    """
    Create a new Rune client session.

    This is the recommended way to initialize a new Rune session for reading
    patient data. Calling `runeq.session()` will initialize a client using
    your local configuration in `~/.rune/config`. This is the same as calling
    `runeq.session(runeq.Config())`.

    config
        (optional) If provided a `runeq.Config`, then the specified
        configuration is used.

    caching
        Set to `False` to disabling local caching of patient data. Disabling
        can impact performance for repeated queries over the same data.
        Default is `True`.

    **kwarg
        When `config` is omitted, you can also pass the configuration options
        directly as keyword parameters. This is shorthand for
        `runeq.session(runeq.Config(**kwarg))`.

    """
    beta_warning()

    if config is None:
        config = Config(**kwarg)
    elif isinstance(config, str):
        config = Config(config)

    return ClientSession(config=config, caching=caching)



"""
Tests for patient device queries.

"""
from unittest import TestCase
from unittest.mock import MagicMock, patch

import runeq


class TestDevices(TestCase):
    """
    Tests for interacting with the patient device graph resource.

    """


    @patch('runeq.v2sdk.client.GraphAPIClient')
    def test_find_by_device_type(self, graphclient):
        """
        Test searching for devices matching a device type.

        """
        session = runeq.session(MagicMock())
        graphclient().fetch_patient.return_value = {
            'id': 'patient-llama0,patient',
            'codeName': 'Llama Zero',
            'createdAt': 1234567890.0
        }
        graphclient().list_patient_devices.side_effect = lambda **kwarg: iter(
            [
                {
                    'id': 'patient-llama0,device-0',
                    'alias': 'Device Zero',
                    'createdAt': 1234567890.1,
                    'deviceType': {
                        'id': 'llamadev0',
                        'displayName': 'Llama Device Type 0'
                    }
                },
                {
                    'id': 'patient-llama0,device-1',
                    'alias': 'Device One',
                    'createdAt': 1234567890.2,
                    'deviceType': {
                        'id': 'llamadev1',
                        'displayName': 'Llama Device Type 1'
                    }
                },
                {
                    'id': 'patient-llama0,device-2',
                    'alias': 'Device Two',
                    'createdAt': 1234567890.2,
                    'deviceType': {
                        'id': 'llamadev0',
                        'displayName': 'Llama Device Type 0'
                    }
                }
            ]
        )

        patient = session.patients['llama0']
        devices = list(patient.devices.find_all_by(device_type='llamadev0'))

        self.assertEqual(2, len(devices))
        self.assertEqual('Device Zero', devices[0].alias)
        self.assertEqual(
            'Llama Device Type 0',
            devices[0].device_type.display_name,
        )
        self.assertEqual('Device Two', devices[1].alias)
        self.assertEqual(
            'Llama Device Type 0',
            devices[1].device_type.display_name,
        )


    @patch('runeq.v2sdk.client.GraphAPIClient')
    def test_frozen_devices(self, graphclient):
        """
        Test listing devices when session is time frozen.

        """
        session = runeq.session(MagicMock())

        graphclient().list_patients.side_effect = lambda: iter(
            [
                {
                    'id': 'patient-llama0,patient',
                    'codeName': 'Llama Zero',
                    'createdAt': 1234567890.0
                }
            ]
        )
        graphclient().list_patient_devices.side_effect = [
            iter([
                {
                    'id': 'patient-llama0,device-0',
                    'alias': 'Device Zero',
                    'createdAt': 1234567890.1,
                    'deviceType': {
                        'id': 'llamadev',
                        'displayName': 'Llama Device'
                    }
                },
                {
                    'id': 'patient-llama0,device-1',
                    'alias': 'Device One',
                    'createdAt': 1234567890.2,
                    'deviceType': {
                        'id': 'llamadev',
                        'displayName': 'Llama Device'
                    }
                }
            ])
        ]

        session.freeze_time(1234567890.15)
        devices = list(session.devices)

        self.assertEqual(1, len(devices))
        self.assertEqual('0', devices[0].id)


    @patch('runeq.v2sdk.client.GraphAPIClient')
    def test_list_all_devices(self, graphclient):
        """
        Test listing all devices across patients

        """
        session = runeq.session(MagicMock())

        graphclient().list_patients.side_effect = lambda: iter(
            [
                {
                    'id': 'patient-llama0,patient',
                    'codeName': 'Llama Zero',
                    'createdAt': 1234567890.0
                },
                {
                    'id': 'patient-llama1,patient',
                    'codeName': 'Llama One',
                    'createdAt': 1234567890.1
                }
            ]
        )
        graphclient().list_patient_devices.side_effect = [
            iter([
                {
                    'id': 'patient-llama0,device-0',
                    'alias': 'Device Zero',
                    'createdAt': 1234567890.2,
                    'deviceType': {
                        'id': 'llamadev',
                        'displayName': 'Llama Device'
                    },
                    'patientId': 'llama0'
                }
            ]),
            iter([
                {
                    'id': 'patient-llama1,device-1',
                    'alias': 'Device One',
                    'createdAt': 1234567890.3,
                    'deviceType': {
                        'id': 'llamadev',
                        'displayName': 'Llama Device'
                    },
                    'patientId': 'llama1'
                },
                {
                    'id': 'patient-llama1,device-2',
                    'alias': 'Device Two',
                    'createdAt': 1234567890.4,
                    'deviceType': {
                        'id': 'llamadev',
                        'displayName': 'Llama Device'
                    },
                    'patientId': 'llama1'
                }
            ])
        ]

        devices = list(session.devices)

        self.assertEqual(3, len(devices))
        self.assertEqual('0', devices[0].id)
        self.assertEqual('llama0', devices[0].patient_id)
        self.assertEqual('1', devices[1].id)
        self.assertEqual('llama1', devices[1].patient_id)
        self.assertEqual('2', devices[2].id)
        self.assertEqual('llama1', devices[2].patient_id)


    @patch('runeq.v2sdk.client.GraphAPIClient')
    def test_list_patient_devices(self, graphclient):
        """
        Test listing a patient's devices.

        """
        session = runeq.session(MagicMock())
        graphclient().fetch_patient.return_value = {
            'id': 'patient-llama0,patient',
            'codeName': 'Llama Zero',
            'createdAt': 1234567890.0
        }
        graphclient().list_patient_devices.side_effect = lambda **kwarg: iter(
            [
                {
                    'id': 'patient-llama0,device-0',
                    'alias': 'Device Zero',
                    'createdAt': 1234567890.1,
                    'deviceType': {
                        'id': 'llamadev',
                        'displayName': 'Llama Device'
                    }
                },
                {
                    'id': 'patient-llama0,device-1',
                    'alias': 'Device One',
                    'createdAt': 1234567890.2,
                    'deviceType': {
                        'id': 'llamadev',
                        'displayName': 'Llama Device'
                    }
                }
            ]
        )

        patient = session.patients['llama0']
        devices = list(patient.devices)

        self.assertEqual(2, len(devices))
        self.assertEqual('0', devices[0].id)
        self.assertEqual('Device Zero', devices[0].alias)
        self.assertEqual('1', devices[1].id)
        self.assertEqual('Device One', devices[1].alias)



"""
Tests for the v2 common library.

"""
from textwrap import dedent
from unittest import TestCase

from runeq.v2sdk.common import SetMemberBase


class TestSetMemberBase(TestCase):
    """
    Tests for the resource collection member class.

    """

    maxDiff = None


    def test_nested_model(self):
        """
        Test modeling a nested structure with relations.

        """
        class Bar(SetMemberBase):
            pass

        class Baz(SetMemberBase):
            pass

        class Foo(SetMemberBase):
            _relations = {
                'bar': Bar,
                'baz': Baz
            }

        foo = Foo({
            'id': 'foo-llama',
            'fieldOne': 'value1',
            'bar': {
                'id': 'bar-llamabar,bar',
                'fieldTwo': 'value2'
            },
            'baz': [
                {
                    'fieldThree': 'value3'
                },
                {
                    'fieldThree': 'value4'
                }
            ]
        })

        self.assertEqual('foo', foo.resource)
        self.assertEqual('llama', foo.id)
        self.assertEqual('value1', foo.field_one)
        self.assertEqual('value1', foo['fieldOne'])
        self.assertEqual('llamabar', foo.bar.id)
        self.assertEqual('value2', foo.bar.field_two)
        self.assertEqual('value3', foo.baz[0].field_three)
        self.assertEqual('value4', foo.baz[1].field_three)

        self.assertEqual(
            dedent(
                '''
                foo {
                    id: llama
                    field_one: value1
                    bar: bar {
                        id: llamabar
                        field_two: value2
                    }
                    baz: [
                        baz {
                            field_three: value3
                        }
                        baz {
                            field_three: value4
                        }
                    ]
                }
                '''
            ).strip(),
            str(foo)
        )



"""
Tests for the Graph API client.

"""
from unittest import TestCase
from unittest.mock import MagicMock, call, patch

from gql import gql

from runeq.config import AUTH_METHOD_ACCESS_TOKEN, AUTH_METHOD_CLIENT_KEYS
from runeq.v2sdk.graph.client import GraphAPIClient
from runeq.v2sdk.graph.model import GraphID



class TestGraphID(TestCase):
    """
    Test the graph identifier wrapper.

    """


    def test_from_absolute(self):
        """
        Test parsing an absolute graph ID.

        """
        graphid = GraphID('patient-llama,device-123')

        self.assertEqual('patient-llama', graphid.principal)
        self.assertEqual('device-123', graphid.relative)
        self.assertEqual('123', graphid.unqualified)
        self.assertEqual('patient-llama,device-123', str(graphid))


    def test_from_relative(self):
        """
        Test constructing from a relative ID.

        """
        graphid = GraphID('llama', 'patient')

        self.assertEqual('patient-llama', graphid.principal)
        self.assertEqual('patient', graphid.relative)
        self.assertEqual('llama', graphid.unqualified)
        self.assertEqual('patient-llama,patient', str(graphid))


    def test_ambigous(self):
        """
        Test that ambiguous IDs raise an error.

        """
        self.assertRaises(ValueError, lambda: GraphID('llama'))



class TestGraphAPIClient(TestCase):
    """
    Tests for the Graph API client.

    """


    @patch('runeq.v2sdk.graph.client.GQLClient')
    def test_fetch_patient(self, gqlclient):
        """
        Test query for single patient by ID.

        """
        config = MagicMock()
        client = GraphAPIClient(config)

        patient0 = {
            'id': 'patient-llama,patient',
            'codeName': 'Llama Patient',
            'createdAt': 1234567890.0
        }
        gqlclient().execute.return_value = {'patient': patient0}

        self.assertEqual(
            patient0,
            client.fetch_patient(GraphID('llama', 'patient'))
        )

        gqlclient().execute.assert_called_once_with(
            gql('''
                query($patientId: ID!) {
                    patient(id: $patientId) {
                        id
                        codeName
                        createdAt
                    }
                }
            '''),
            variable_values={'patientId': 'patient-llama,patient'}
        )


    @patch('runeq.v2sdk.graph.client.GQLClient')
    def test_list_patient_devices(self, gqlclient):
        """
        Test query for listing devices for a patient.

        """
        config = MagicMock()
        client = GraphAPIClient(config)

        device0 = {
            'id': 'patient-llama,device-0',
            'alias': "Device Zero",
            'createdAt': 1234567890.0,
            'deviceType': {
                'id': 'llama',
                'displayName': "Llama Device"
            }
        }
        device1 = {
            'id': 'patient-llama,device-1',
            'alias': "Device One",
            'createdAt': 1234567890.1,
            'deviceType': {
                'id': 'llama',
                'displayName': "Llama Device"
            }
        }
        device2 = {
            'id': 'patient-llama,device-2',
            'alias': "Device Two",
            'createdAt': 1234567890.2,
            'deviceType': {
                'id': 'llama',
                'displayName': "Llama Device"
            }
        }

        gqlclient().execute.side_effect = [
            {
                'patient': {
                    'deviceList': {
                        'pageInfo': {
                            'endCursor': 'llamacursor0',
                        },
                        'devices': [
                            device0,
                            device1,
                        ]
                    }
                }
            },
            {
                'patient': {
                    'deviceList': {
                        'pageInfo': {
                            'endCursor': None,
                        },
                        'devices': [
                            device2
                        ]
                    }
                }
            }
        ]

        devices = client.list_patient_devices(
            patient_id=GraphID('patient-llama,patient')
        )

        self.assertEqual(device0, next(devices))
        self.assertEqual(device1, next(devices))
        self.assertEqual(device2, next(devices))
        self.assertRaises(StopIteration, lambda: next(devices))

        expected_gql = gql('''
            query(
                $patientId: ID!,
                $withDisabled: Boolean!,
                $cursor: Cursor
            ) {
                patient(id: $patientId) {
                    deviceList(
                        withDisabled: $withDisabled,
                        cursor: $cursor
                    ) {
                        devices {
                            id
                            alias
                            createdAt
                            deviceType {
                                id
                                displayName
                            }
                        }
                        pageInfo {
                            endCursor
                        }
                    }
                }
            }
        ''')

        self.assertEqual(2, gqlclient().execute.call_count)
        gqlclient().execute.assert_has_calls([
            call(
                expected_gql,
                variable_values={
                    'cursor': None,
                    'patientId': 'patient-llama,patient',
                    'withDisabled': False
                }
            ),
            call(
                expected_gql,
                variable_values={
                    'cursor': 'llamacursor0',
                    'patientId': 'patient-llama,patient',
                    'withDisabled': False
                }
            )
        ])


    @patch('runeq.v2sdk.graph.client.GQLClient')
    def test_list_patients(self, gqlclient):
        """
        Test query for listing all patients.

        """
        config = MagicMock()
        client = GraphAPIClient(config)

        patient0 = {
            'id': 'patient-llama0,patient',
            'codeName': 'Patient Llama 0',
            'createdAt': 1234567890.0,
        }
        patient1 = {
            'id': 'patient-llama1,patient',
            'codeName': 'Patient Llama 1',
            'createdAt': 1234567890.1,
        }
        patient2 = {
            'id': 'patient-llama2,patient',
            'codeName': 'Patient Llama 2',
            'createdAt': 1234567890.2,
        }

        gqlclient().execute.side_effect = [
            {
                'org': {
                    'patientAccessList': {
                        'pageInfo': {
                            'endCursor': 'llamacursor0',
                        },
                        'patientAccess': [
                            {'patient': patient0},
                            {'patient': patient1}
                        ]
                    }
                }
            },
            {
                'org': {
                    'patientAccessList': {
                        'pageInfo': {
                            'endCursor': None,
                        },
                        'patientAccess': [
                            {'patient': patient2}
                        ]
                    }
                }
            }
        ]

        patients = client.list_patients()

        self.assertEqual(patient0, next(patients))
        self.assertEqual(patient1, next(patients))
        self.assertEqual(patient2, next(patients))
        self.assertRaises(StopIteration, lambda: next(patients))

        expected_gql = gql('''
            query($cursor: Cursor) {
                org {
                    patientAccessList(cursor: $cursor) {
                        pageInfo {
                            endCursor
                        }
                        patientAccess {
                            patient {
                                id
                                codeName
                                createdAt
                            }
                        }
                    }
                }
            }
        ''')

        self.assertEqual(2, gqlclient().execute.call_count)
        gqlclient().execute.assert_has_calls([
            call(expected_gql, variable_values={'cursor': None}),
            call(expected_gql, variable_values={'cursor': 'llamacursor0'})
        ])


    @patch('runeq.v2sdk.graph.client.GQLClient')
    def test_whoami_patient(self, gqlclient):
        """
        Test that the "who am I?" query for a patient.

        """
        config = MagicMock(auth_method=AUTH_METHOD_CLIENT_KEYS)
        client = GraphAPIClient(config)

        gqlclient().execute.return_value = {
            'patient': {
                'id': 'patient-llama,patient',
                'codeName': 'Patient Llama',
                'createdAt': 1234567890.0,
            }
        }

        self.assertEqual(
            gqlclient().execute.return_value,
            client.whoami()
        )

        gqlclient().execute.assert_called_once_with(
            gql('''
                query {
                    patient {
                        id
                        codeName
                        createdAt
                    }
                }
            '''),
            variable_values={}
        )


    @patch('runeq.v2sdk.graph.client.GQLClient')
    def test_whoami_user(self, gqlclient):
        """
        Test that the "who am I?" query for a user.

        """
        config = MagicMock(auth_method=AUTH_METHOD_ACCESS_TOKEN)
        client = GraphAPIClient(config)

        gqlclient().execute.return_value = {
            'user': {
                'id': 'llama',
                'created': 1234567890.0,
                'defaultMembership': {
                    'id': 'llama',
                    'created': 1234567890.0,
                    'org': {
                        'id': 'llama',
                        'created': 1234567890.0,
                        'displayName': 'Llama Org',
                    }
                },
                'displayName': 'Llama User',
                'email': 'llama@runelabs.io',
                'username': 'llamauser'
            }
        }

        self.assertEqual(
            gqlclient().execute.return_value,
            client.whoami()
        )

        gqlclient().execute.assert_called_once_with(
            gql('''
                query {
                    user {
                        id
                        created
                        defaultMembership {
                            id
                            created
                            org {
                                id
                                created
                                displayName
                            }
                        }
                        displayName
                        email
                        username
                    }
                }
            '''),
            variable_values={}
        )



"""
Models for querying patient metadata.

"""
from typing import Iterable

from .common import SetMemberBase, QueryBase, SetBase
from .devices import DeviceQuery, DeviceSet
from .graph.model import GraphID


def _patient_graph_id(patient_id: str) -> str:
    """
    Qualify a patient ID for use in the Graph API.

    """
    if '-' not in patient_id:
        patient_id = f"patient-{patient_id}"

    if ',' not in patient_id:
        patient_id = f"{patient_id},patient"

    return patient_id


class Patient(SetMemberBase):
    """
    Representational state for a patient.

    """


    def __init__(self, res, session):
        """
        Initialize a patient representation, connected to a session for
        further active record querying.

        """
        super().__init__(res)
        self._session = session
        self._device_cache = DeviceSet()


    @property
    def devices(self) -> DeviceQuery:
        """
        Begin a query over this patient's registered devices.

        """
        return DeviceQuery(session=self._session, patient=self)


class PatientSet(SetBase):
    """
    Representational state for a collection of patients.

    """

    _member = Patient


    def __getitem__(self, patient_id: str) -> Patient:
        """
        Get a patient by ID.

        """
        return super().__getitem__(_patient_graph_id(patient_id))


class PatientsQuery(QueryBase):
    """
    A query over registered patients.

    """

    _set = PatientSet


    def __getitem__(self, patient_id: str) -> Patient:
        """
        Fetch a patient by ID.

        """
        sess = self._session

        try:
            return sess._patient_cache[patient_id]

        except KeyError:
            patient = Patient(
                sess.graph.fetch_patient(GraphID(patient_id, 'patient')),
                session=sess
            )
            sess._patient_cache.add(patient)
            return patient


    def __iter__(self):
        """
        Execute the query and return an iterator over the results.

        """
        sess = self._session

        if sess._caching:
            #
            # Use patient cache
            #
            if not sess._patient_cache._all:
                sess._patient_cache = self.query()

            if sess._now is not None:
                yield from sess._patient_cache.created_before(sess._now)
            else:
                yield from iter(sess._patient_cache)

        else:
            #
            # Iterate over paginated set, only buffering one page at a time
            #
            for patient_state in self.raw_query():
                patient = Patient(patient_state, session=sess)

                if sess._now is not None and patient.created_at >= sess._now:
                    continue

                yield patient


    def raw_query(self) -> Iterable:
        """
        Query patient state from the API.

        """
        return self._session.graph.list_patients()
